//*****************************************************************
//* PCX to Neo Geo tile converter.
//*****************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>

#define PAL_SIZE        16
#define TILEDIMENSION	16

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef BYTE           BOOL;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

int  Rip(char *filename);// function declarations                
void WriteTiles(void);
int  LoadPCX(const char* file, BYTE *outpix, BYTE * pal);
long FileSize(const char *fname);
void GetTile(BYTE *tile, int x, int y, int width, int height);

#define BANKSIZE   (1024 * 1024)

FILE   *fp;
FILE   *stream;              // level output data file
WORD    size = 0;
int     Width, Height, pal, start;
WORD    palval;
FILE	*handle;
char	filebank1[32];
char	filebank2[32];
char	filepal[32];
BYTE    bank1[BANKSIZE];
BYTE    bank2[BANKSIZE];
WORD	palette[PAL_SIZE];
long	offset	= 0;

typedef struct PCXHeader
{
    BYTE id;
    BYTE version;
    BYTE rle;
    BYTE bpp;
    WORD xstart;                         
    WORD ystart;
    WORD xend;
    WORD yend;
    WORD hres;
    WORD vres;
    BYTE pal[48];
    BYTE rsvd1;
    BYTE nbitp;
    WORD bytesperline;
    WORD paltype;
    WORD hsize;
    WORD vsize;
    BYTE rsvd2[54];
} PCX;

#define TILESIZE  256     // tiles are 16 pixels x 16 pixels

BYTE PCXBuffer[64000];
BYTE PCXPalette[768];

void Usage();

void main(int argc, char *argv[])
{
    printf("PCX to Neo Geo tile converter.\n");

    if (argc < 5)
    {
        Usage();
    }

    memset(bank1, 0, BANKSIZE);
    memset(bank2, 0, BANKSIZE);
    memset(palette, 0, PAL_SIZE * 2);
    
    offset	= atol(argv[1]);

	strcpy(filebank1, argv[3]);

    if ("" == filebank1)
    {
        Usage();
    }

    strcpy(filebank2, argv[4]);

    if ("" == filebank2)
    {
        Usage();
    }

	if (offset > 0)
	{
		handle = fopen(filebank1, "rb");

		if (handle != NULL)
		{
			fread(bank1, 1, BANKSIZE, handle);
			fclose(handle);
		}

		handle = fopen(filebank2, "rb");

		if (handle != NULL)
		{
			fread(bank2, 1, BANKSIZE, handle);
			fclose(handle);
		}
	}

    if (argc > 5)
    {
        strcpy(filepal, argv[5]);

        if (strlen(filepal) <= 0)
        {
            printf("Palette will not be saved.\n");

            goto  EnterWidth;
        }
    }

    else
    {
        memset(filepal, 0, sizeof(filepal));
    }

EnterWidth:
    if (-1 == Rip(argv[2]))
    {
        exit(1);
    }

    WriteTiles();

	printf("Next Offset = %d", offset); 
}

void Usage()
{
    printf("Command syntax is: TileNG Offset Infile Rom1 Rom2 [Palette]\n");

    exit(1);
}

//**************************** Rip *******************************
int Rip(char *filename)
{
    int  x, y;
	BYTE	*Plane1[2];
	BYTE	*Plane2[2];
	BYTE	*Plane3[2];
	BYTE	*Plane4[2];

    LoadPCX(filename, PCXBuffer, PCXPalette);

    if ((Width % TILEDIMENSION) != 0 && (Height % TILEDIMENSION) != 0)
    {
        printf("Picture must be a multiple of %d.\n", TILEDIMENSION);
        
		return -1;
    }

    BYTE	Tile[TILESIZE];
	Plane1[0]	= bank2 + 1 + offset;
    Plane2[0]	= bank2 + offset;
    Plane3[0]	= bank1 + 1 + offset;
    Plane4[0]	= bank1 + offset;
    Plane1[1]	= bank2 + 33 + offset;
    Plane2[1]	= bank2 + 32 + offset;
    Plane3[1]	= bank1 + 33 + offset;
    Plane4[1]	= bank1 + 32 + offset;

    for (int yloop = 0; yloop < (Height / TILEDIMENSION); yloop++)
    {
        for (int xloop = 0; xloop < (Width / TILEDIMENSION); xloop++)
        {
            x	= xloop * TILEDIMENSION;
            y	= yloop * TILEDIMENSION;

            GetTile(&Tile[0], x, y, TILEDIMENSION, TILEDIMENSION);

            for (int tiley = 0; tiley < TILEDIMENSION; tiley++)
            {
                *Plane1[0]	= 0;
                *Plane2[0]	= 0;
                *Plane3[0]	= 0;
                *Plane4[0]	= 0;
                *Plane1[1]	= 0;
                *Plane2[1]	= 0;
                *Plane3[1]	= 0;
                *Plane4[1]	= 0;
				
				for (int tilex = 0; tilex < TILEDIMENSION / 2; tilex++)
                {
					BYTE Pixel1	= Tile[tiley * TILEDIMENSION + tilex];
					BYTE Pixel2	= Tile[tiley * TILEDIMENSION + tilex + 8];
				
					*Plane4[0]	|= ((Pixel2 & 1) >> 0) << (tilex);
					*Plane3[0]	|= ((Pixel2 & 2) >> 1) << (tilex);
					*Plane2[0]	|= ((Pixel2 & 4) >> 2) << (tilex);
					*Plane1[0]	|= ((Pixel2 & 8) >> 3) << (tilex);
					
					*Plane4[1]	|= ((Pixel1 & 1) >> 0) << (tilex);
					*Plane3[1]	|= ((Pixel1 & 2) >> 1) << (tilex);
					*Plane2[1]	|= ((Pixel1 & 4) >> 2) << (tilex);
					*Plane1[1]	|= ((Pixel1 & 8) >> 3) << (tilex);
				}

				Plane1[0]	+= 2;
				Plane2[0]	+= 2;
				Plane3[0]	+= 2;
				Plane4[0]	+= 2;
								
				Plane1[1]	+= 2;				
				Plane2[1]	+= 2;				
				Plane3[1]	+= 2;				
				Plane4[1]	+= 2;				
			}
        
			Plane1[0]	+= 32;
			Plane2[0]	+= 32;
			Plane3[0]	+= 32;
			Plane4[0]	+= 32;
								
			Plane1[1]	+= 32;				
			Plane2[1]	+= 32;				
			Plane3[1]	+= 32;				
			Plane4[1]	+= 32;				
		}
    }

	offset	= (long)(Plane2[0] - bank2);
    
	return 0;
}

void WriteTiles(void)
{
    for (int palloop = 0; palloop < PAL_SIZE; palloop++)
    {
        BYTE red    = ((PCXPalette[palloop * 3]) >> 2);
        BYTE green  = ((PCXPalette[palloop * 3 + 1]) >> 2);
        BYTE blue   = ((PCXPalette[palloop * 3 + 2]) >> 2);

        red   &= 0x0f;
        green &= 0x0f;
        blue  &= 0x0f;

		WORD	value = ((WORD)green << 12) | ((WORD)blue << 8) | ((WORD)red << 0);
        palette[palloop] = value;	
    }

    stream = fopen(filepal, "wb");

	if (NULL == stream)
	{
		return;
	}

	fwrite(palette, 2, sizeof(palette) / sizeof(WORD), stream);
	fclose(stream);

    // open a binary file for the level array
    stream = fopen(filebank1, "wb");

	if (NULL == stream)
	{
		return;
	}

    fwrite(bank1, 1, sizeof(bank1), stream);

    fclose(stream);

    // open a binary file for the level array
    stream = fopen(filebank2, "wb");

	if (NULL == stream)
	{
		return;
	}

    fwrite(bank2, 1, sizeof(bank2), stream);

    fclose(stream);
}

int LoadPCX(const char* file, BYTE *outpix, BYTE * pal)
{
    int         i, w, h;
    PCXHeader   hdr;
    long        pos;
    int         ret = 0;

    fp = fopen(file, "rb");

    if (fp == NULL)
        goto bye;

    pos = ftell(fp);
    fread (&hdr, 1, sizeof(hdr), fp);

    if (hdr.id != 0x0A || hdr.version != 5 || hdr.rle != 1
     || (hdr.bpp == 8 && hdr.nbitp != 1))
        goto bye;

    w       = hdr.xend - hdr.xstart + 1;
    Width   = w;

    h       = hdr.yend - hdr.ystart + 1;
    Height  = h;

    if (hdr.bpp == 1 && hdr.nbitp == 4)
    {
         memset(pal, 0, 768);

         for (i = 0; i < 48; i++)
            pal[i] = hdr.pal[i] >> 2;
    }

    else if (hdr.bpp == 8 && hdr.nbitp == 1)
    {
        fseek(fp, pos + FileSize(file) - 768, SEEK_SET);

        for (i = 0; i < 768; i++)
            pal[i] = ((BYTE)getc(fp)) >> 2;
    }

    else
        goto bye;

    ret = 1;

    fseek(fp, pos + sizeof(PCXHeader), SEEK_SET);

    while (h-- > 0)
    {
        BYTE   c;
        BYTE  *outpt;
        int             np = 0;

        outpt = outpix;

        memset(outpix, 0, w);

        for (np = 0; np < hdr.nbitp; np++)
        {
            i = 0;
            outpix = outpt;

            do
            {
                c = (BYTE)getc(fp);

                if ((c & 0xC0) != 0xC0)
                {
                    if (hdr.bpp == 1)
                    {
                        int k;
                        for (k = 7; k >= 0; k--)
                            *outpix++ |= ((c >> k) & 1) << np;
                        i += 8;
                    }

                    else
                    {
                        *outpix++ = c;
                        i++;
                    }
                }

                else
                {
                    BYTE v;
                    v = (BYTE)getc(fp);
                    c &= ~0xC0;

                    while (c > 0 && i < w)
                    {
                        if (hdr.bpp == 1)
                        {
                            int k;

                            for (k = 7; k >= 0; k--)
                                *outpix++ |= ((v >> k) & 1) << np;

                            i += 8;
                        }

                        else
                        {
                            *outpix++ = v;
                            i++;
                        }

                        c--;
                    }
                }
            }
            while (i < w);
        }
    }

  bye:
    if (fp != NULL)
        fclose(fp);

    return ret;
}

long FileSize(const char *fname)
{
    FILE *f;
    long length;

    f = fopen(fname, "rb");

    if (f == NULL)
        return 0;

    fseek(f, 0, SEEK_END);

    length = ftell(f);

    fclose(f);

    return length;
}

void GetTile(BYTE *tile, int x, int y, int width, int height)
{
    BYTE *localtile = tile;
    BYTE *buffer    = &PCXBuffer[0];

    buffer += y * Width + x;

    for (int outerloop = 0; outerloop < height; outerloop++)
    {
        for (int innerloop = 0; innerloop < width; innerloop++)
        {
            *localtile++ = *buffer++;
        }

        buffer += Width - width;
    }
}
